@@ -10,7 +10,6 @@ import (
 	"regexp"
 	"slices"
 	"strings"
-	"unicode"
 
 	"github.com/alecthomas/chroma/v2/quick"
 	"github.com/apex/log"
@@ -59,6 +58,13 @@ func (i *Imports) uniq(foundation map[string][]string) {
 	i.Locals = slices.Compact(i.Locals)
 	i.Classes = slices.Compact(i.Classes)
 	i.Protos = slices.Compact(i.Protos)
+	i.Imports = slices.DeleteFunc(i.Imports, func(l string) bool {
+		l = strings.TrimSuffix(l, "-Protocol.h")
+		l = strings.TrimSuffix(l, ".h")
+		_, foundC := slices.BinarySearch(foundation["classes"], l)
+		_, foundP := slices.BinarySearch(foundation["protocols"], l)
+		return foundC || foundP
+	})
 	i.Locals = slices.DeleteFunc(i.Locals, func(l string) bool {
 		l = strings.TrimSuffix(l, "-Protocol.h")
 		l = strings.TrimSuffix(l, ".h")
@@ -214,21 +220,23 @@ func (o *ObjC) DumpProtocol(pattern string) error {
 
 		for _, proto := range protos {
 			if re.MatchString(proto.Name) {
-				if o.conf.Color {
-					if o.conf.Addrs {
-						quick.Highlight(os.Stdout, swift.DemangleBlob(proto.WithAddrs()), "objc", "terminal256", o.conf.Theme)
+				if _, ok := seen[proto.Ptr]; !ok { // prevent displaying duplicates
+					if o.conf.Color {
+						if o.conf.Addrs {
+							quick.Highlight(os.Stdout, swift.DemangleBlob(proto.WithAddrs()), "objc", "terminal256", o.conf.Theme)
+						} else {
+							quick.Highlight(os.Stdout, swift.DemangleBlob(proto.Verbose()), "objc", "terminal256", o.conf.Theme)
+						}
+						quick.Highlight(os.Stdout, "\n/****************************************/\n\n", "objc", "terminal256", o.conf.Theme)
 					} else {
-						quick.Highlight(os.Stdout, swift.DemangleBlob(proto.Verbose()), "objc", "terminal256", o.conf.Theme)
-					}
-					quick.Highlight(os.Stdout, "\n/****************************************/\n\n", "objc", "terminal256", o.conf.Theme)
-				} else {
-					if o.conf.Addrs {
-						fmt.Println(swift.DemangleBlob(proto.WithAddrs()))
-					} else {
-						fmt.Println(swift.DemangleBlob(proto.Verbose()))
+						if o.conf.Addrs {
+							fmt.Println(swift.DemangleBlob(proto.WithAddrs()))
+						} else {
+							fmt.Println(swift.DemangleBlob(proto.Verbose()))
+						}
 					}
+					seen[proto.Ptr] = true
 				}
-				seen[proto.Ptr] = true
 			}
 		}
 	}
@@ -883,9 +891,11 @@ func (o *ObjC) processForwardDeclarations(m *macho.File) (map[string]Imports, er
 	}
 
 	for _, class := range classes {
-		imp := Imports{}
-		if class.SuperClass != "NSObject" { // skip NSObject since we'll import Foundation by default
-			imp.Imports = append(imp.Imports, class.SuperClass+".h")
+		var imp Imports
+		if slices.Contains(classNames, class.SuperClass) {
+			imp.Locals = append(imp.Locals, class.SuperClass+".h")
+		} else {
+			imp.Classes = append(imp.Classes, class.SuperClass)
 		}
 		for _, prot := range class.Protocols {
 			if slices.Contains(protoNames, prot.Name) {
@@ -896,76 +906,16 @@ func (o *ObjC) processForwardDeclarations(m *macho.File) (map[string]Imports, er
 		}
 		for _, ivar := range class.Ivars {
 			typ := ivar.Type
-			if strings.ContainsAny(typ, "<>") {
-				typ = strings.Trim(typ, "@\"")
-				if rest, ok := strings.CutPrefix(typ, "NSObject<"); ok {
-					typ = strings.TrimSuffix(rest, ">")
-					if slices.Contains(protoNames, typ) {
-						imp.Locals = append(imp.Locals, typ+"-Protocol.h")
-					} else {
-						imp.Protos = append(imp.Protos, typ)
-					}
-				}
-				typ = strings.Trim(typ, "<>")
-				if slices.Contains(protoNames, typ) {
-					imp.Locals = append(imp.Locals, typ+"-Protocol.h")
-				} else {
-					imp.Protos = append(imp.Protos, typ)
-				}
-
-			} else {
-				if rest, ok := strings.CutPrefix(typ, "@\""); ok {
-					typ = strings.TrimSuffix(rest, "\"")
-					if slices.Contains(classNames, typ) {
-						imp.Locals = append(imp.Locals, typ+".h")
-					} else {
-						imp.Classes = append(imp.Classes, typ)
-					}
-				}
-			}
+			o.fillImportsForType(typ, class.Name, classNames, protoNames, &imp)
 		}
 		for _, prop := range class.Props {
 			typ := prop.Type()
-			if strings.ContainsAny(typ, "<>") {
-				typ = strings.Trim(typ, "@\"")
-				typ = strings.Trim(typ, " *")
-				if rest, ok := strings.CutPrefix(typ, "NSObject<"); ok {
-					typ = strings.TrimSuffix(rest, ">")
-					if slices.Contains(protoNames, typ) {
-						imp.Locals = append(imp.Locals, typ+"-Protocol.h")
-					} else {
-						imp.Protos = append(imp.Protos, typ)
-					}
-				}
-				typ = strings.Trim(typ, "<>")
-				if slices.Contains(protoNames, typ) {
-					imp.Locals = append(imp.Locals, typ+"-Protocol.h")
-				} else {
-					imp.Protos = append(imp.Protos, typ)
-				}
-			} else {
-				if (len(typ) > 0 && unicode.IsUpper(rune(typ[0])) && unicode.IsLetter(rune(typ[0]))) && strings.HasSuffix(typ, "*") {
-					typ = strings.Trim(typ, " *")
-					if slices.Contains(classNames, typ) {
-						imp.Locals = append(imp.Locals, typ+".h")
-					} else {
-						imp.Classes = append(imp.Classes, typ)
-					}
-				}
-			}
+			o.fillImportsForType(typ, class.Name, classNames, protoNames, &imp)
 		}
 		for _, method := range class.InstanceMethods {
 			for i := 0; i < method.NumberOfArguments(); i++ {
 				typ := method.ArgumentType(i)
-				if (len(typ) > 0 && unicode.IsUpper(rune(typ[0])) && unicode.IsLetter(rune(typ[0]))) && strings.HasSuffix(typ, "*") { // or < >
-					if slices.Contains(classNames, typ) {
-						imp.Locals = append(imp.Locals, typ+".h")
-					} else if slices.Contains(protoNames, strings.Trim(typ, "NSObject<>")) {
-						imp.Locals = append(imp.Locals, strings.Trim(typ, "NSObject<>")+"-Protocol.h")
-					} else {
-						imp.Classes = append(imp.Classes, typ)
-					}
-				}
+				o.fillImportsForType(typ, class.Name, classNames, protoNames, &imp)
 				if i == 0 {
 					i += 2
 				}
@@ -974,13 +924,7 @@ func (o *ObjC) processForwardDeclarations(m *macho.File) (map[string]Imports, er
 		for _, method := range class.ClassMethods {
 			for i := 0; i < method.NumberOfArguments(); i++ {
 				typ := method.ArgumentType(i)
-				if (len(typ) > 0 && unicode.IsUpper(rune(typ[0])) && unicode.IsLetter(rune(typ[0]))) && strings.HasSuffix(typ, "*") { // or < >
-					if slices.Contains(classNames, typ) {
-						imp.Locals = append(imp.Locals, typ+".h")
-					} else {
-						imp.Classes = append(imp.Classes, typ)
-					}
-				}
+				o.fillImportsForType(typ, class.Name, classNames, protoNames, &imp)
 				if i == 0 {
 					i += 2
 				}
@@ -997,7 +941,7 @@ func (o *ObjC) scanFoundation() error {
 	o.foundation["classes"] = []string{}
 	o.foundation["protocols"] = []string{}
 	if o.cache != nil {
-		for _, name := range []string{"Foundation", "CoreFoundation"} {
+		for _, name := range []string{"Foundation", "CoreFoundation", "libobjc.A.dylib"} {
 			img, err := o.cache.Image(name)
 			if err != nil {
 				return err
@@ -1038,3 +982,77 @@ func (o *ObjC) scanFoundation() error {
 	}
 	return nil
 }
+
+func (o *ObjC) fillImportsForType(typ string, className string, classNames []string, protoNames []string, imp *Imports) {
+	if o.isBuiltInType(typ) {
+		return
+	}
+
+	typ = strings.Trim(typ, ` "*@^`)
+
+	if !strings.Contains(typ, "<") {
+		if !slices.Contains(classNames, typ) {
+			imp.Classes = append(imp.Classes, typ)
+			return
+		}
+
+		if typ != className {
+			imp.Locals = append(imp.Locals, typ+".h")
+			return
+		}
+
+		return
+	}
+
+	typ = strings.ReplaceAll(typ, "><", ", ")
+
+	if !strings.HasPrefix(typ, "<") {
+		before, after, _ := strings.Cut(typ, "<")
+
+		o.fillImportsForType(before, className, classNames, protoNames, imp)
+		o.fillImportsForType("<"+after, className, classNames, protoNames, imp)
+
+		return
+	}
+
+	typ = strings.Trim(typ, "<>")
+
+	for _, typ := range strings.Split(typ, ", ") {
+		if !slices.Contains(protoNames, typ) {
+			imp.Protos = append(imp.Protos, typ)
+			continue
+		}
+
+		imp.Locals = append(imp.Locals, typ+"-Protocol.h")
+	}
+}
+
+func (o *ObjC) isBuiltInType(typ string) bool {
+	if typ == "" {
+		return true
+	}
+
+	if strings.ContainsAny(typ, "#(:?[{") {
+		return true
+	}
+
+	switch typ[0] {
+	case '"', '@', '^':
+		return o.isBuiltInType(typ[1:])
+	}
+
+	switch typ {
+	case "*", "B", "C", "I", "L", "N", "O", "Q", "R", "S", "V", "b", "c", "d", "f", "i", "l", "n", "o", "q", "r", "s", "v":
+		return true
+	}
+
+	switch before, after, _ := strings.Cut(typ, " "); before {
+	case "_Atomic", "const", "restrict", "volatile":
+		return o.isBuiltInType(after)
+
+	case "_Bool", "_Complex", "_Imaginary", "BOOL", "Class", "IMP", "Ivar", "Method", "SEL", "bool", "char", "class", "double", "enum", "float", "id", "int", "long", "short", "signed", "struct", "union", "unsigned", "void":
+		return true
+	}
+
+	return false
+}
